#include "libs/cit_comm.h"
#include "libs/cit_common.h"

#include "bsp/hybrid_events.h"
#include "bsp/nrf5340dk_pins.h"
#include "utils/crc8.h"
#include "libs/cit_common.h"

#include <FreeRTOS.h>
#include <nrfx_spim.h>
#include <nrfx_spis.h>

#include <string.h>
#include <stdio.h>

#include "bsp/debug_rtt.h"

#define KHZ(x) (1000 * x)

#define CIT_BUF_SIZE 55
#define CIT_RX_BUFFERS 32
#define CIT_WRAP_MASK 0x1F   // Note: keep in  sync with CIT_RX_BUFFERS

typedef struct {
    struct {
        nrfx_spim_t instance;
        htimer_t event;
        uint8_t tx_buf[CIT_BUF_SIZE];
        uint8_t rx_buf[CIT_BUF_SIZE];
        volatile uint8_t busy;  // 0 = not busy, 1=busy, 2=delay for spi
    } send;

    struct {
        nrfx_spis_t instance;

        // use a ring-buffer for frames and one big array for data
        volatile uint16_t head;  // write-index
        volatile uint16_t tail;  // read-index
        cit_comm_frame_t frames[CIT_RX_BUFFERS];
        uint8_t rx_buf[CIT_RX_BUFFERS][CIT_BUF_SIZE];
        uint8_t tx_buf[CIT_BUF_SIZE];

        uint16_t min_buffers;   // for checking if buffersize is ok
        uint32_t overruns;
        uint32_t spi_errors;
        uint32_t rx_count;
    } recv;

    cit_config_t cbs;
} cit_t;

static cit_t cit = {
    .send = {
        .instance = NRFX_SPIM_INSTANCE(3),
    },
    .recv = {
        .instance = { NRF_SPIS2, 0 },
    },
};

static void on_spim_event_irq(const nrfx_spim_evt_t* event, void* context)
{
    if (event->type == NRFX_SPIM_EVENT_DONE) {
        htimer_add_irq(&cit.send.event); // schedule timer for mainloop
    }
}

static inline uint16_t cit_comm_get_count(void) {
    uint16_t tail = cit.recv.tail;
    uint16_t head = cit.recv.head;

    if (head >= tail) return head - tail;

    return CIT_RX_BUFFERS - (tail - head);
}

static inline uint16_t cit_comm_get_free_count(void) {
    return CIT_RX_BUFFERS - cit_comm_get_count() - 1;
}

static void set_buffers(void)
{
    cit.recv.rx_buf[cit.recv.head][0] = 0;  // set first byte to 0

    nrfx_err_t err = nrfx_spis_buffers_set(&cit.recv.instance,
                                           cit.recv.tx_buf, CIT_BUF_SIZE,
                                           cit.recv.rx_buf[cit.recv.head], CIT_BUF_SIZE);
    my_assert(err == NRFX_SUCCESS, "SPIS setbuffer");
}

static void spis_event_irq(const nrfx_spis_evt_t* event, void* context)
{
    if (event->evt_type == NRFX_SPIS_XFER_DONE) {

        cit.recv.rx_count++;
        cit.recv.frames[cit.recv.head].size = event->rx_amount;

        if (event->rx_amount == CIT_COMM_REPLY_LEN || event->rx_amount == CIT_COMM_MAX_DATA_LEN) {
            uint16_t num_free = cit_comm_get_free_count();
            if (num_free < cit.recv.min_buffers) {
                cit.recv.min_buffers = num_free;
            }
            if (num_free == 0) {
                cit.recv.overruns++;
                cit.recv.min_buffers = CIT_RX_BUFFERS-1;
                // Note: if no space, overwrite current
            } else {
                cit.recv.head = (cit.recv.head + 1) & CIT_WRAP_MASK;
            }
            set_buffers();
            cit.cbs.on_recv_irq();   // also notify if buffers full
        } else {
            debug_rtt_printf("SPI ERR (len %d)\n", event->rx_amount);
            cit.recv.spi_errors++;
            set_buffers();
        }
    }
}

static void cit_on_idle_timer(void* arg)
{
    if (cit.send.busy == 1) {
        cit.send.busy = 2;
        htimer_add_ticks(&cit.send.event, 2);   // delay 2 ticks (=shortest reliable interval)
    } else {
        cit.send.busy = 0;
        cit.cbs.on_idle();
    }
}

void cit_comm_init(const cit_config_t* config)
{
    htimer_assign(&cit.send.event, cit_on_idle_timer, NULL, "cit_timer");

    memcpy(&cit.cbs, config, sizeof(cit_config_t));
    for (uint16_t i = 0; i < CIT_RX_BUFFERS; i++) {
        cit.recv.frames[i].data = cit.recv.rx_buf[i];
    }
    cit.recv.min_buffers = CIT_RX_BUFFERS-1;

    // SPI bus1: send bus
    nrfx_spim_config_t spim_config = {
        .sck_pin  = PIN_HDR7,
        .mosi_pin = PIN_HDR8,
        .miso_pin = NRF_SPIM_PIN_NOT_CONNECTED,
        .ss_pin   = PIN_HDR9,
        .ss_active_high = false,
        .irq_priority = 7,
        .orc = 0xFF,
        .frequency = KHZ(500),
        .mode = NRF_SPIM_MODE_0,
        .bit_order = NRF_SPIM_BIT_ORDER_MSB_FIRST,
        .miso_pull = NRF_GPIO_PIN_PULLUP,
        .skip_gpio_cfg = false,
        .skip_psel_cfg = false,
    };

    nrfx_err_t res = nrfx_spim_init(&cit.send.instance, &spim_config, on_spim_event_irq, NULL);
    my_assert(res == NRFX_SUCCESS, "spim init");

    // SPI bus2: receive bus
    nrfx_spis_config_t spis_config = {
        .sck_pin  = PIN_HDR4,
        .mosi_pin = PIN_HDR5,
        .miso_pin = NRF_SPIM_PIN_NOT_CONNECTED,
        .csn_pin  = PIN_HDR6,
        .mode = NRF_SPIS_MODE_0,
        .bit_order = NRF_SPIS_BIT_ORDER_MSB_FIRST,
        .csn_pullup = NRF_GPIO_PIN_PULLUP,
        .miso_drive = NRF_GPIO_PIN_S0S1,
        .def = 0xFF,
        .orc = 0xFF,
        .irq_priority = 7,
        .skip_gpio_cfg = false,
        .skip_psel_cfg = false,
    };

    nrfx_err_t err = nrfx_spis_init(&cit.recv.instance, &spis_config, spis_event_irq, NULL);
    my_assert(err == NRFX_SUCCESS, "spis init");

    set_buffers();
}

bool cit_comm_is_tx_busy(void)
{
    return (cit.send.busy != 0);
}

void cit_comm_send(const uint8_t* data, uint32_t len)
{
    my_assert(cit.send.busy == 0, "CIT send busy");
    uint8_t* tx_buf = cit.send.tx_buf;
    memcpy(tx_buf, data, len);

    cit.send.busy = 1;
    nrfx_spim_xfer_desc_t xfer_desc = { tx_buf, len, cit.send.rx_buf, len };
    nrfx_err_t res = nrfx_spim_xfer(&cit.send.instance, &xfer_desc, 0);
    if (res != NRFX_SUCCESS) {
        debug_rtt_writestring("spim transfer failed\n");
    }
}

cit_comm_frame_t* cit_comm_get_frame(void)
{
    if (cit.recv.tail == cit.recv.head) return NULL;

    return &cit.recv.frames[cit.recv.tail];
}

void cit_comm_consume_frame(void)
{
    // Note: doesn't check anything
    cit.recv.tail = (cit.recv.tail + 1) & CIT_WRAP_MASK;
}

